/*   acua_acctmod - Allows a user to modify their acua account.
 *   Copyright (C) 2000 Robert Davidson.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   Written by Robert Davidson, E-Mail: acua@acua.ebbs.com.au
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>
#include <sys/types.h>          /*  waitpid  */
#include <sys/wait.h>           /*  waitpid  */
#include <sys/socket.h>         /*  inet_   */
#include <netinet/in.h>         /*  inet_   */
#include <arpa/inet.h>          /*  inet_   */
#include <time.h>
#include <cgi.h>                /* cgilib */
#include "radutmp.h"

#define VERSION		"ACUA Account Mod CGI Version: 0.01"

#ifndef RADUTMP
#define RADUTMP "/var/log/radutmp"
#endif

/* Define this if you want logging enabled. Logging is done via syslog()
 * Set the below to be "1", otherwise "0" to not log anything.
 */
#define LOGGING 1
#define DEBUG 1

/* Files that this program uses */
#define ACUA_PATH               "/usr/bin/acua"
#define FILE_BASE               "/usr/lib/acua/acua_acctmod"
#define FILE_BAD_USAGE		FILE_BASE"/bad_usage.html"
#define FILE_CHANGE_OK_E	FILE_BASE"/success_e.html"
#define FILE_CHANGE_OK_D	FILE_BASE"/success_d.html"
#define FILE_CHANGE_NO		FILE_BASE"/fail.html"
#define FILE_INIT_FAIL          FILE_BASE"/init_fail.html"
#define FILE_MISC_ERR		FILE_BASE"/misc_err.html"
#define FILE_NO_LOGIN           FILE_BASE"/no_login.html"
#define FILE_MAY_FAIL           FILE_BASE"/may_fail.html"
#define FILE_HTML_HEADER        FILE_BASE"/header.html"
#define FILE_HTML_FOOTER        FILE_BASE"/footer.html"

/* end of configurable stuff */

/* -----------------------------------------------------------------------
 * - Start Of Program. ---------------------------------------------------
 * -----------------------------------------------------------------------
 */

/* Echo a file to stdout */
int display_file(char *fname)
{
    int         c;
    FILE        *fp;

    fp = fopen(fname, "r");
    if (!fp) return(-1);
    while (!feof(fp))
    {
	c = fgetc(fp);
	if (c == EOF) break;
	fputc(c, stdout);
    }
    return(0);
}

void display_html_header(void)
{
    if (display_file(FILE_HTML_HEADER))
	printf("<HTML>\n<BODY>\n<PRE>\n");
    return;
}

void display_html_footer(void)
{
    if (display_file(FILE_HTML_FOOTER))
        printf("</PRE>\n</BODY>\n</HTML>\n");
    return;
}

void segv_handler(int aaa)
{
    syslog(LOG_ERR, "Segmentation fault caught.");
    closelog();
    exit(1);
}

/* magically called when the CGI exits */
void clean_exit(void)
{
    display_html_footer();
    closelog();
    fflush(stdout);
    return;
}                                        

int main(int argc, char **argv)
{

    int		r_addr;                 /* Remote IP Address in NBO. */
    int         c_pid;                  /* Child pid. */
    int         c_wdog;                 /* Child watchdog timer. */
    int         c_stat;                 /* Child status. */
    int         action;                 /* Action requested. */
    char        username[RUT_NAMESIZE]; /* Username of requesting user. */
    /* this is so that the version is readable by the strings program. */
    char	acctmod_version[64] = VERSION;
    FILE        *fp;
    struct      in_addr inp;
    struct      radutmp rt;

    /* CGI Library Variables */
    s_cgi       *cgi_data;

    /* Segfault handler for debugging */
    signal(SIGSEGV, segv_handler);

    /* Open the syslog */
    if (LOGGING) openlog("acua_acctmod", LOG_PID, LOG_USER);

    /* call clean_exit before terminating */
    atexit(clean_exit);

    /* set debugging level and initialize the cgi library */
    cgiDebug(0, 0);
    if (!(cgi_data = cgiInit()))
    {
	/* Failed for some reason to initialize the cgi library */
	if (LOGGING)
	    syslog(LOG_ERR, "Failed to initialize CGI library");
        display_html_header();
	if (display_file(FILE_INIT_FAIL))
	    printf("<A>Failed to initialize the CGI library.</A>\n");
        exit(1);
    }

    /* Print out the cgi header */
    cgiHeader();        		/* Content-type: text/html */

    /* Check that argc is correct */
    if (argc > 1)
    {
	if (display_file(FILE_BAD_USAGE))
            printf("<A>Incorrect CGI usage.  Use POST method.</A>\n");
	exit(1);
    }

    /* Are we running as the root user? We need to, so that we can modify
     * users acua accounts.
     */
    if (geteuid() != 0)
    {
	printf("<A>Please make this CGI setuid root.</A>\n");
	syslog(LOG_ERR, "This cgi program needs to be setuid root");
        exit(1);
    }

    /* Figure out the IP Address of the remote host */
    r_addr = inet_aton(getenv("REMOTE_ADDR"), &inp);
    if (!r_addr)
    {
	if (display_file(FILE_MISC_ERR))
	    printf("<A>There was a problem determining your IP Address</A>\n");
	if (LOGGING)
	    syslog(LOG_ERR, "%s", "Could not determine IP of remote client.");
	exit(1);
    }

    /* Log where the request came from */
    if (LOGGING)
	syslog(LOG_INFO, "Connect from: [%s]", inet_ntoa(inp));

    /* From their IP Address, figure out who the user is */
    if (DEBUG)
	syslog(LOG_ERR, "Opening RADUTMP File");
    username[0]='\0';
    if (!(fp = fopen(RADUTMP, "r"))) {
	if (LOGGING)
	    syslog(LOG_ERR, "Could not open file for reading: %s", RADUTMP);
	if (display_file(FILE_MISC_ERR))
            printf("<A>A server-side error has occoured.</A>\n");
	exit(1);
    }
    if (DEBUG)
	syslog(LOG_ERR, "Opened RADUTMP File");
    /* read and compare the logins */
    while(fread(&rt, sizeof(rt), 1, fp) == 1)
    {
	/* Skip to next record if this login is not active */
	if (rt.type != 1) continue;

        /* Is the ip address the same? */
	if (inp.s_addr == rt.framed_address)
	{
	    strncpy(username, rt.login, RUT_NAMESIZE);
	    break;
	}
    }

    /* Check to see if we found them or not */
    if (username[0] == '\0')
    {
	if (display_file(FILE_NO_LOGIN))
	    printf("<A>Sorry, you do not appear to be logged in at the moment.</A>\n");
        exit(1);
    }
    syslog(LOG_ERR, "Username: [%s]", username);

    /* The client is now authenticated */

    /* Validate the form that was posted */
    if (!cgiGetValue(cgi_data, "overtime"))
	if (display_file(FILE_BAD_USAGE))
	{
	    printf("<A>Form param 'overtime' is not complete.</A>\n");
	    exit(1);
	}

    action = strtol(cgiGetValue(cgi_data, "overtime"), (char **)NULL, 10);
    if ((errno == ERANGE) || (action > 1) || (action < 0))
    {
	if (display_file(FILE_BAD_USAGE))
	    printf("<A>Form param 'overtime' is out of range.</A>\n");
	exit(1);
    }

    /* Form 'overtime' param has passed the sanity checks */

    if (action == 0)
    {
	/* Turn off the overtime feature, it's sloppy, but executing an
	 * ACUA program looks easier than linking with libacua and all
	 * that crap, plus it's already been bug tested heaps :)
	 */
	c_pid = fork();

	if (c_pid == 0)
	{
	    /* This is the child process */
	    execl(ACUA_PATH, "acua", "modRec", username, "-OVERTIME", NULL);
	    if (LOGGING)
		syslog(LOG_ERR, "%s", "An error occoured while running acua modRec");
            exit(1);
	}
    }

    if (action == 1)
    {
	/* Turn on the overtime feature */
	c_pid = fork();

	if (c_pid == 0)
	{
	    /* This is the child process */
	    execl(ACUA_PATH, "acua", "modRec", username, "+OVERTIME", NULL);
	    if (LOGGING)
		syslog(LOG_ERR, "%s", "An error occoured while running acua modRec");
            exit(1);
	}
    }

    /* Parent process - check for errors, wait for the child to exit */
    if (c_pid == -1)
    {
	if (display_file(FILE_CHANGE_NO))
	    printf("<A>An error occoured while enabling the feature.</A>\n");
	if (LOGGING)
	    syslog(LOG_ERR, "%s", "Could not fork new process, change failed");
	exit(1);
    }

    c_wdog = time(NULL);
    c_stat = 0;
    for (;;) {
	if (waitpid(c_pid, &c_stat, WNOHANG) > 0) break;
	if (time(NULL) > c_wdog+10) {
	    if (display_file(FILE_CHANGE_NO))
		printf("<A>An error occoured while processing your request.</A>\n");
	    if (LOGGING)
		syslog(LOG_ERR, "%s", "Warning: child crashed, killing!");
	    kill(c_pid, SIGKILL); /* go away */
	    exit(1);
	}
	usleep(250);
    }

    /* examine the c_stat value to determine if it was a success */
    if (WIFEXITED(c_stat))
    {
	if (WEXITSTATUS(c_stat) == 1)
	{
	    /* execl probably failed */
	    if (display_file(FILE_CHANGE_NO))
		printf("<A>Child process failed.</A>\n");
	    if (LOGGING)
		syslog(LOG_ERR, "Child process failed");
	    exit(1);
	}

	if (WEXITSTATUS(c_stat) == 0)
	{
	
	    if (action == 0)
	    {
		if (display_file(FILE_CHANGE_OK_D))
	        printf("<A>The feature you requested has been disabled.</A>\n");
	        exit(0);
	    }

	    if (action == 1)
	    {
		if (display_file(FILE_CHANGE_OK_E))
	        printf("<A>The feature you requested has been enabled.</A>\n");
		exit(0);
	    }

	    printf("<A>Code error or some knit wit is playing with the CGI</A>\n");
	    exit(1);
        }
    }

    /* Users should never see this error, hopefully! */
    if (display_file(FILE_MAY_FAIL))
	printf("<A>Your request may have failed.</A>\n");

    return(1);
}
